@@ -1,5 +1,5 @@
-import json
 import io
+import json
 import os
 from functools import wraps
 from typing import Any, Dict, List, Optional
@@ -194,10 +194,8 @@ class HdDB:
         :raises ConnectionError: Si hay un error al eliminar la tabla de Motherduck
         """
         try:
-
             self.execute("BEGIN TRANSACTION;")
             try:
-
                 self.execute(f'DROP TABLE IF EXISTS "{org}__{db}"."{tbl}";')
 
                 self.execute(f'DELETE FROM "{org}__{db}".hd_tables WHERE id = ?', [tbl])
@@ -207,7 +205,6 @@ class HdDB:
 
                 self.execute("COMMIT;")
             except Exception as e:
-
                 self.execute("ROLLBACK;")
                 raise e
 
@@ -474,20 +471,22 @@ class HdDB:
             self.execute("ROLLBACK;")
             logger.error(f"Error adding column to table {tbl}: {e}")
             raise QueryError(f"Error adding column to table {tbl}: {e}")
-        
+
     @attach_motherduck
     def delete_column(self, org: str, db: str, tbl: str, column: dict):
         try:
             column_name = column["slug"]
             column_id = column["fld___id"]
             self.execute("BEGIN TRANSACTION;")
-            self.execute(f'ALTER TABLE "{org}__{db}"."{tbl}" DROP COLUMN "{column_name}"')
+            self.execute(
+                f'ALTER TABLE "{org}__{db}"."{tbl}" DROP COLUMN "{column_name}"'
+            )
             self.execute(
                 f'UPDATE "{org}__{db}".hd_tables SET ncol = ncol - 1 WHERE id = ?',
                 [tbl],
             )
             self.execute(
-                f'DELETE FROM "{org}__{db}".hd_fields WHERE id = ? AND tbl = ?',
+                f'DELETE FROM "{org}__{db}".hd_fields WHERE fld___id = ? AND tbl = ?',
                 [column_id, tbl],
             )
             self.execute("COMMIT;")
@@ -496,7 +495,86 @@ class HdDB:
             self.execute("ROLLBACK;")
             logger.error(f"Error deleting column from table {tbl}: {e}")
             raise QueryError(f"Error deleting column from table {tbl}: {e}")
-    
+
+    @attach_motherduck
+    def upload_bulk_data(self, org: str, db: str, tbl: str, data: pd.DataFrame):
+        try:
+            # Fetch the column names and their order from the actual table
+            table_columns = self.execute(f'DESCRIBE "{org}__{db}"."{tbl}"').fetchall()
+            table_column_names = [col[0] for col in table_columns]
+
+            # Fetch the column labels and ids from hd_fields
+            hd_fields = self.execute(
+                f'SELECT id, label FROM "{org}__{db}".hd_fields WHERE tbl = ?',
+                [tbl],
+            ).fetchall()
+            hd_fields_dict = {field[1]: field[0] for field in hd_fields}
+            hd_fields_reverse_dict = {field[0]: field[1] for field in hd_fields}
+
+            # Create a mapping from data column names to table column names
+            column_mapping = {}
+            for data_col in data.columns:
+                if data_col in hd_fields_dict:
+                    column_mapping[data_col] = hd_fields_dict[data_col]
+                else:
+                    column_mapping[data_col] = data_col  # Keep unmapped columns as is
+
+            logger.info(f"column_mapping: {column_mapping}")
+
+            # Check if all required columns are present in the data and identify extra columns
+            missing_columns = set(table_column_names) - set(column_mapping.values())
+            extra_columns = set(column_mapping.values()) - set(table_column_names)
+
+            if missing_columns or extra_columns:
+                error_message = []
+                if missing_columns:
+                    missing_labels = [
+                        hd_fields_reverse_dict.get(col, col) for col in missing_columns
+                    ]
+                    error_message.append(
+                        f"Missing columns in data: {', '.join(missing_labels)}"
+                    )
+                if extra_columns:
+                    extra_labels = [
+                        hd_fields_reverse_dict.get(col, col) for col in extra_columns
+                    ]
+                    error_message.append(
+                        f"Extra columns in data: {', '.join(extra_labels)}"
+                    )
+                raise ValueError(". ".join(error_message))
+
+            self.execute("BEGIN TRANSACTION;")
+
+            # Reorder and rename the columns in the DataFrame to match the table structure
+            data_reordered = data.rename(columns=column_mapping)[table_column_names]
+
+            # Insert data into the table
+            self.execute(
+                f'INSERT INTO "{org}__{db}"."{tbl}" SELECT * FROM data_reordered'
+            )
+
+            # Get the number of rows inserted
+            rows_inserted = len(data)
+
+            # Update the number of rows in hd_tables
+            self.execute(
+                f"""
+                UPDATE "{org}__{db}".hd_tables 
+                SET nrow = nrow + ? 
+                WHERE id = ?
+            """,
+                [rows_inserted, tbl],
+            )
+
+            self.execute("COMMIT;")
+            return True
+        except duckdb.Error as e:
+            self.execute("ROLLBACK;")
+            logger.error(f"Error uploading bulk data to table {tbl}: {e}")
+            raise QueryError(f"Error uploading bulk data to table {tbl}: {e}")
+        except ValueError as e:
+            logger.error(str(e))
+            raise QueryError(str(e))
 
     @attach_motherduck
     def update_hd_fields(self, org: str, db: str, fld___id: str, label: str, type: str):
@@ -506,3 +584,18 @@ class HdDB:
         except duckdb.Error as e:
             logger.error(f"Error updating hd_fields: {e}")
             raise QueryError(f"Error updating hd_fields: {e}")
+
+    @attach_motherduck
+    def update_row(self, org: str, db: str, tbl: str, id: str, data: dict):
+        try:
+            # Construct the UPDATE query
+            update_query = f'UPDATE "{org}__{db}".{tbl} SET '
+            for key in data.keys():  # Removed value
+                update_query += f"{key} = ?, "
+            update_query = update_query.rstrip(", ") + " WHERE rcd___id = ?"
+
+            # Execute the UPDATE query
+            self.execute(update_query, list(data.values()) + [id])
+        except duckdb.Error as e:
+            logger.error(f"Error updating row in table {tbl}: {e}")
+            raise QueryError(f"Error updating row in table {tbl}: {e}")
